# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cache
from mio import cfg
from mio import eal

import subprocess


def gen_doxygen(ip_str):
    vendor, name = common.parse_dep(ip_str)
    if vendor == "":
        ip = cache.get_anon_ip(name, True)
    else:
        ip = cache.get_ip(vendor, name, True)
    ip_name = f"{ip.vendor}/{ip.name}"
    ip_src_path = ip.path + "/" + ip.src_path
    ip_bin_path = ip.path + "/" + ip.scripts_path
    args  = " PROJECT_NAME='"  + ip.full_name + "'"
    args += " PROJECT_BRIEF='" + ip.full_name + "'"
    args += " IP_NAME='" + ip.name + "'"
    args += " PROJECT_NUMBER='" + ip.name + "'"
    args += " EXAMPLE_PATH=" + ip.path + "/" + ip.examples_path
    args += " OUTPUT_PATH="  + ip.path + "/" + ip.docs_path + "/dox_out"
    args += " SRC_PATH=" + ip_src_path + " MIO_HOME=" + cfg.mio_data_src_dir + " IP_NAME=" + ip_name
    
    args += " DOCS_PATH+=" + ip.path + "/" + ip.docs_path
    #args += " IMAGE_PATH=" + ip_path + "/" + ip_metadata['structure']['docs-path']
    #args += " INPUT+="     + ip_path + "/" + ip_metadata['structure']['docs-path']
    #for input_dir in ip_metadata['hdl-src']['directories']:
    #    if input_dir != "." and input_dir != "":
    #        args += " INPUT+=" + ip_src_path + "/" + input_dir
    #    else:
    #        args += " INPUT+=" + ip_src_path
    common.info(f"Invoking Doxygen on IP '{ip_name}' ({ip_src_path})")
    eal.launch_eda_bin(args + " doxygen", [cfg.mio_data_src_dir + "/doxygen.private.cfg"], cfg.mio_data_dir)
    common.info("Done.  To view documentation: `firefox " + ip.path + "/" + ip.docs_path + "/dox_out/html/index.html &`")



# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################




from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import cache
from mio import common
from mio import eal
from mio import install
from mio import doctor
from tqdm import tqdm
from tqdm import trange
from threading import Thread
from multiprocessing.pool import ThreadPool
from threading import BoundedSemaphore
import threading
import time
import os
from datetime import datetime
import yaml
from yaml.loader import SafeLoader
import math
import re
import atexit
import tarfile




bwrap_ignore_list = [
    "xsim.dir", ".str", ".Xil", ".jou", ".log", ".wdb", ".vcd", ".log", ".sdb", ".rlx", ".pb", ".o", ".png", ".jpg",
    ".svg", ".vsdx", ".docx", ".xlsx", ".pptx", ".md", "sync", "workspace"
]

bwrap_ignore_dirs = [ ".git", ".svn" ]

regex_define_pattern  = "\+define\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
regex_plusarg_pattern = "\+((?:\w|_|\d)+)(?:\=((?:\w|_|\d)+))?"
seconds_waited = 0
num_deps_to_compile = 0
sem = BoundedSemaphore(1)
pbar = None



class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self,  *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()




class SimulationJob:
    """Simulation Job model"""
    
    def __init__(self, ip_str):
        self.vendor, self.ip = common.parse_dep(ip_str)
        self.simulator       = ""
        self.one_shot        = True
        self.fsoc            = False
        self.compile         = False
        self.elaborate       = False
        self.simulate        = False
        self.test            = ""
        self.seed            = 0
        self.max_errors      = 0
        self.gui             = False
        self.verbosity       = ""
        self.waves           = False
        self.cov             = False
        self.dry_run         = False
        self.raw_args        = []
        self.cmp_args        = {}
        self.elab_args       = {}
        self.sim_args        = {}
        
        self.bwrap           = False
        self.bwrap_commands  = []
        self.bwrap_flists    = {}
        
        self.is_regression        = False
        self.regression_name      = ""
        self.regression_timestamp = ""
        
        self.timestamp_start    = ""
        self.timestamp_end      = ""
        self.filelist_path      = ""
        self.results_path       = ""
        self.results_dir_name   = ""
        self.cmp_log_file_path  = ""
        self.elab_log_file_path = ""
        self.sim_log_file_path  = ""
        




def main(sim_job):
    global est_time
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg(f"Starting simulation job: '{sim_job.vendor}/{sim_job.ip}'")
    if sim_job.vendor == "":
        ip = cache.get_anon_ip(sim_job.ip, True)
    else:
        ip = cache.get_ip(sim_job.vendor, sim_job.ip, True)
    if ip == None:
        common.fatal(f"Cannot find IP '{sim_job.vendor}/{sim_job.ip}'")
    ip_str = f"{ip.vendor}/{ip.name}"
    
    if sim_job.compile:
        if not ip.are_deps_installed():
            deps_to_install = ip.get_deps_to_install()
            if len(deps_to_install) > 0:
                install_deps_str = ""
                while (install_deps_str != "y") and (install_deps_str != "n"):
                    install_deps_str = common.prompt(f"{len(deps_to_install)} dependencies must first be installed.  Would you like to do so now? [y/n]").strip().lower()
                if install_deps_str == "y":
                    local_str = ""
                    while (local_str != "y") and (local_str != "n"):
                        local_str = common.prompt("Local install (vs. global)? [y/n]").strip().lower()
                    if local_str == "y":
                        global_install = False
                    else:
                        global_install = True
                    common.info(f"Installing {len(deps_to_install)} dependencies ...")
                    while len(deps_to_install) > 0:
                        try:
                            install.install_ip_dep_list(ip, deps_to_install, global_install)
                        except Exception as e:
                            common.fatal("Error during '" + ip_str + "' IP installation: " + str(e))
                        cache.scan_and_load_ip_metadata()
                        deps_to_install = ip.get_deps_to_install()
                else:
                    common.fatal("Cannot continue without first installing '" + ip_str + "' IP dependencies.")
    
    if not doctor.check_simulator_executables(sim_job.simulator):
        common.fatal("Simulator '" + sim_job.simulator + "' not installed properly or environment variable missing")
    
    if sim_job.simulator == common.simulators_enum.METRICS:
        eal.init_metrics_workspace()
    
    convert_cli_args_to_defines (sim_job)
    convert_cli_args_to_plusargs(sim_job)
    create_sim_directories()
    
    check_dependencies(ip)
    ip.update_is_compiled_elaborated(sim_job.simulator)
    
    if sim_job.one_shot:
        one_step_sim(sim_job)
    else:
        multi_step_sim(sim_job)
    
    if sim_job.bwrap:
        bubble_wrap(sim_job)


def one_step_sim(sim_job):
    global est_time
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.vendor == "":
        ip = cache.get_anon_ip(sim_job.ip, True)
    else:
        ip = cache.get_ip(sim_job.vendor, sim_job.ip, True)
    if ip == None:
        common.fatal(f"Cannot find IP '{sim_job.vendor}/{sim_job.ip}'")
    ip_str = f"{ip.vendor}/{ip.name}"
    flist_path = ""
    fsoc_core_name = ""
    if ip.dut_ip_type == "fsoc":
        if not ip.dut_core.is_installed:
            flist_path = eal.invoke_fsoc(ip, ip.dut_core, sim_job)
            ip.dut_core.is_installed = True
            fsoc_core_name = ip.dut_core
        else:
            common.info("Skipping processing of DUT FuseSoC core '" + ip.dut_fsoc_full_name + "'.")
    else:
        if ip.dut != None:
            if ip.dut.target_ip_model == None:
                common.fatal(f"Did not resolve DUT dependency ('{ip.dut.vendor}/{ip.dut.target_ip}')!")
            if ip.dut.target_ip_model.sub_type == "vivado":
                common.fatal("Vivado Project DUTs are not yet amenable to single-step compilation/elaboration flow.")
    
    common.info(f"Compiling+Elaborating {ip_str} ...")
    eal.gen_ip_image(ip, sim_job, fsoc_core_name, flist_path)
    if not sim_job.is_regression:
        common.banner(f"Simulating {ip_str} ...")
        eal.simulate(ip, sim_job)
        print_end_of_simulation_message(ip, sim_job)


def multi_step_sim(sim_job):
    global est_time
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.vendor == "":
        ip = cache.get_anon_ip(sim_job.ip, True)
    else:
        ip = cache.get_ip(sim_job.vendor, sim_job.ip, True)
    if ip == None:
        common.fatal(f"Cannot find IP '{sim_job.vendor}/{sim_job.ip}'")
    ip_str = f"{ip.vendor}/{ip.name}"
    
    cmp_ip_count = 0
    if sim_job.compile:
        if ip.has_dut:
            compile_dut = True
            if ip.dut_ip_type == "fsoc":
                compile_dut = True #ip.dut_core.is_compiled[sim_job.simulator]
            else:
                compile_dut = True #ip.is_compiled[sim_job.simulator]
            if compile_dut:
                cmp_ip_count = cmp_ip_count + 1
                if ip.dut_ip_type == "fsoc":
                    dut_str = f"{ip.dut_fsoc_name}"
                else:
                    dut_str = f"{ip.dut.vendor}/{ip.dut.target_ip}"
                if (not sim_job.dry_run) and (dut_str in cfg.job_history):
                    est_time = 0
                    curr_est_time = 0
                    if 'compilation' in cfg.job_history[dut_str]:
                        for job in cfg.job_history[dut_str]['compilation']:
                            start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                            end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                            curr_est_time = end - start
                            curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                            est_time += curr_est_time
                        est_time = math.ceil(est_time / len(cfg.job_history[dut_str]['compilation']))
                        if est_time > 0:
                            pool = ThreadPool(processes=1)
                            pool.apply_async(progress_bar)
                            cmp_dut(ip, sim_job)
                            pool.terminate()
                            pool.join()
                        else:
                            cmp_dut(ip, sim_job)
                    else:
                        cmp_dut(ip, sim_job)
                else:
                    cmp_dut(ip, sim_job)
        
        cmp_ip_count = cmp_ip_count + cmp_dependencies(ip, sim_job)
        
        #if not ip.is_compiled[sim_job.simulator]:
        if True:
            if not sim_job.is_regression:
                common.banner("Compiling IP '" + ip_str + "'")
            cmp_ip_count = cmp_ip_count + 1
            if (not sim_job.dry_run) and (ip_str in cfg.job_history):
                if 'compilation' in cfg.job_history[ip_str]:
                    est_time = 0
                    curr_est_time = 0
                    for job in cfg.job_history[ip_str]['compilation']:
                        start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                        end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                        curr_est_time = end - start
                        curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                        est_time += curr_est_time
                    est_time = math.ceil(est_time / len(cfg.job_history[ip_str]['compilation']))
                    if est_time > 0:
                        pool = ThreadPool(processes=1)
                        pool.apply_async(progress_bar)
                        cmp_target_ip(ip, sim_job)
                        pool.terminate()
                        pool.join()
                    else:
                        cmp_target_ip(ip, sim_job)
                else:
                    cmp_target_ip(ip, sim_job)
            else:
                cmp_target_ip(ip, sim_job)
    
    needs_elaboration = True #not ip.is_elaborated[sim_job.simulator]
    #if cmp_ip_count > 0:
    #    needs_elaboration = True
    
    if sim_job.elaborate and needs_elaboration:
        if sim_job.is_regression:
            common.info("Elaborating IP '" + ip_str + "'")
        else:
            common.banner("Elaborating IP '" + ip_str + "'")
        est_time = 0
        curr_est_time = 0
        if ip_str in cfg.job_history:
            if (not sim_job.dry_run) and ('elaboration' in cfg.job_history[ip_str]):
                for job in cfg.job_history[ip_str]['elaboration']:
                    start = datetime.strptime(job['timestamp_start'], "%Y/%m/%d-%H:%M:%S")
                    end   = datetime.strptime(job['timestamp_end'  ], "%Y/%m/%d-%H:%M:%S")
                    curr_est_time = end - start
                    curr_est_time = divmod(curr_est_time.seconds, 60)[1]
                    est_time += curr_est_time
                est_time = math.ceil(est_time / len(cfg.job_history[ip_str]['elaboration']))
                if est_time > 0:
                    pool = ThreadPool(processes=1)
                    pool.apply_async(progress_bar)
                    eal.elaborate(ip, sim_job)
                    pool.terminate()
                    pool.join()
                else:
                    eal.elaborate(ip, sim_job)
            else:
                eal.elaborate(ip, sim_job)
        else:
            eal.elaborate(ip, sim_job)
    else:
        if sim_job.simulate == False and not sim_job.is_regression:
            print_end_of_compilation_message(ip, sim_job)
    
    if sim_job.simulate:
        if sim_job.gui and sim_job.simulator == common.simulators_enum.METRICS:
            common.warning("The Metrics Cloud Simulator does not support GUI mode")
        if not sim_job.is_regression:
            common.banner("Simulating IP '" + ip_str + "'")
        eal.simulate(ip, sim_job)
        if sim_job.compile == True and not sim_job.is_regression:
            print_end_of_compilation_message(ip, sim_job)
        if sim_job.elaborate == True and not sim_job.is_regression:
            print_end_of_elaboration_message(ip, sim_job)
        if not sim_job.is_regression:
            print_end_of_simulation_message(ip, sim_job)
    else:
        if sim_job.elaborate == True and not sim_job.is_regression:
            if sim_job.compile == True:
                print_end_of_compilation_message(ip, sim_job)
            print_end_of_elaboration_message(ip, sim_job)



def bubble_wrap(sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    run_script_path = f"{cfg.sim_dir}/run.sh"
    readme_path = f"{cfg.sim_dir}/README.txt"
    
    if sim_job.simulator == common.simulators_enum.VIVADO:
        bin_text = "MIO_VIVADO_HOME"
        bin_string = "/tools/vivado/bin"
    if sim_job.simulator == common.simulators_enum.METRICS:
        bin_text = "MIO_METRICS_HOME"
        bin_string = "/usr/local/bin"
    if sim_job.simulator == common.simulators_enum.VCS:
        bin_text = "MIO_VCS_HOME"
        bin_string = "/tools/vcs/bin"
    if sim_job.simulator == common.simulators_enum.QUESTA:
        bin_text = "MIO_QUESTA_HOME"
        bin_string = "/tools/questa/bin"
    if sim_job.simulator == common.simulators_enum.XCELIUM:
        bin_text = "MIO_XCELIUM_HOME"
        bin_string = "/tools/xcelium/bin"
    if sim_job.simulator == common.simulators_enum.RIVIERA:
        bin_text = "MIO_RIVIERA_HOME"
        bin_string = "/tools/riviera/bin"
    
    try:
        with open(run_script_path, 'w') as run_script_file:
            run_script_file.write(f"export PROJECT_ROOT_DIR=$(pwd)/..\n\n")
            for flist in sim_job.bwrap_flists:
                run_script_file.write(f"export {flist}={sim_job.bwrap_flists[flist]}\n")
            run_script_file.write("\n\n\n\n")
            for cmd in sim_job.bwrap_commands:
                cmd = cmd.replace(cfg.project_dir , "${PROJECT_ROOT_DIR}")
                cmd = cmd.replace("-f .mio/"      , "-f ${PROJECT_ROOT_DIR}/.mio/")
                cmd = cmd.replace(cfg.vivado_home , "${MIO_VIVADO_HOME}" )
                cmd = cmd.replace(cfg.metrics_home, "${MIO_METRICS_HOME}")
                cmd = cmd.replace(cfg.vcs_home    , "${MIO_VCS_HOME}"    )
                cmd = cmd.replace(cfg.xcelium_home, "${MIO_XCELIUM_HOME}")
                cmd = cmd.replace(cfg.questa_home , "${MIO_QUESTA_HOME}" )
                cmd = cmd.replace(cfg.riviera_home, "${MIO_RIVIERA_HOME}")
                run_script_file.write(f"{cmd}\n\n")
        common.info(f"Wrote {run_script_path}")
        with open(readme_path, 'w') as readme_file:
            readme_file.write(f"1. Set ${bin_text}.  Ex: export {bin_text}={bin_string}\n")
            readme_file.write(f"2. Run: bash ./run.sh\n")
        common.info(f"Wrote {readme_path}")
        
        tarball_filename = f"{sim_job.ip}.{sim_job.test}.{sim_job.seed}.{sim_str}.tgz"
        tarball_path = f"{cfg.project_dir}/../{tarball_filename}"
        common.info(f"Writing {tarball_path} ...")
        with tarfile.open(tarball_path, "w:gz") as tar:
            files = os.listdir(cfg.project_dir)
            for file in files:
                file_path = f"{cfg.project_dir}/{file}"
                if os.path.isdir(file_path):
                    if file in bwrap_ignore_dirs:
                        continue
                common.dbg(f"Compressing {file}")
                tar.add(file_path, filter=bwrap_exclude, arcname=file)
        common.info("Done")
        
    except Exception as e:
        common.fatal(f"Failed to create bubble-wrap tarball: {e}")



def bwrap_exclude(tar_info):
    for regex in bwrap_ignore_list:
        if tar_info.name.endswith(regex):
            return None
    return tar_info



def kill_progress_bar():
    pass
atexit.register(kill_progress_bar)



def progress_bar():
    global pbar
    global seconds_waited
    with tqdm(total=est_time) as pbar:
        for i in range(est_time):
            sleep(1)
            pbar.update(1)
    #with tqdm(total=seconds) as pbar:
    #    for i in range(seconds):
    #        sleep(1)
    #        pbar.update(1)
        
    #with alive_bar(seconds, bar = 'smooth', stats="{eta} estimated", monitor=False, elapsed=True) as bar:
    #    bar.text("estimated")
    #    for x in range(seconds):
    #        time.sleep(1)
    #        seconds_waited += 1
    #        bar()




def convert_cli_args_to_defines(sim_job):
    defines = {}
    args = sim_job.raw_args
    if args != None:
        if type(args) is list:
            for item in args:
                regex = "\+define\+(\w+)(?:\=(\w+))?"
                result = re.match(regex, item)
                if result:
                    if len(result.groups()) > 1:
                        arg_name  = result.group(1)
                        arg_value = result.group(2)
                        if arg_value == None:
                            arg_value = ""
                    elif len(result.groups()) == 1:
                        arg_name  = result.group(1)
                        arg_value = ""
                    else:
                        continue
                else:
                    continue
                defines[arg_name] = arg_value
                common.dbg("Added define '" + arg_name + "' with value '" + arg_value + "' to list")
        else:
            all_args = re.sub("\"", "", args)
            for arg in all_args:
                result = re.match(regex_define_pattern, arg)
                if result:
                    define_name = result.group(1)
                    if len(result.groups()) > 1:
                        define_value = result.group(2)
                        if define_value == None:
                            arg_value = ""
                    else:
                        define_value = ""
                    defines[define_name] = define_value
                    common.dbg("Added define '" + define_name + "' with value '" + define_value + "' to list")
    sim_job.cmp_args = defines


def convert_cli_args_to_plusargs(sim_job):
    define_regex = "\+define\+(\w+)(?:\=(\w+))?"
    plus_args = {}
    args = sim_job.raw_args
    if args != None:
        if type(args) is list:
            for item in args:
                result = re.match(define_regex, item)
                if result:
                    continue
                regex = "\+(\w+)(?:\=(\w+))?"
                result = re.match(regex, item)
                if result:
                    if len(result.groups()) > 1:
                        arg_name  = result.group(1)
                        arg_value = result.group(2)
                        if arg_value == None:
                            arg_value = ""
                    elif len(result.groups()) == 1:
                        arg_name  = result.group(1)
                        arg_value = ""
                    else:
                        continue
                else:
                    continue
                plus_args[arg_name] = arg_value
                common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
        else:
            all_args = re.sub("\"", "", args)
            for arg in all_args:
                result = re.match(regex_define_pattern, arg)
                if not result:
                    result = re.match(regex_plusarg_pattern, arg)
                    if result:
                        arg_name = result.group(1)
                        if len(result.groups()) > 1:
                            arg_value = result.group(2)
                            if arg_value == None:
                                arg_value = ""
                        else:
                            arg_value = ""
                        plus_args[arg_name] = arg_value
                        common.dbg("Added plus arg '" + arg_name + "' with value '" + arg_value + "' to list")
    sim_job.sim_args = plus_args


def create_sim_directories():
    common.dbg("Creating sim directories")
    common.create_dir(cfg.sim_dir)
    common.create_dir(cfg.sim_output_dir)
    common.create_dir(cfg.sim_output_dir + "/viv"                     )
    common.create_dir(cfg.sim_output_dir + "/viv/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/viv/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/viv/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/viv/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/viv/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/vcs"                     )
    common.create_dir(cfg.sim_output_dir + "/vcs/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/vcs/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/vcs/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/vcs/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/vcs/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/xcl"                     )
    common.create_dir(cfg.sim_output_dir + "/xcl/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/xcl/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/xcl/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/xcl/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/xcl/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/qst"                     )
    common.create_dir(cfg.sim_output_dir + "/qst/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/qst/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/qst/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/qst/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/qst/so_libs"             )
    common.create_dir(cfg.sim_output_dir + "/riv"                     )
    common.create_dir(cfg.sim_output_dir + "/riv/cov_wd"              )
    common.create_dir(cfg.sim_output_dir + "/riv/cmp_out"             )
    common.create_dir(cfg.sim_output_dir + "/riv/sim_wd"              )
    common.create_dir(cfg.sim_output_dir + "/riv/regr_wd"             )
    common.create_dir(cfg.sim_output_dir + "/riv/so_libs"             )
    common.create_dir(cfg.sim_dir + "/cmp")
    common.create_dir(cfg.sim_dir + "/elab")
    common.create_dir(cfg.regr_results_dir)
    common.create_dir(cfg.sim_results_dir)


def check_dependencies(ip):
    for dep in ip.dependencies:
        common.dbg(f"Checking dependency '{dep.vendor}/{dep.target_ip}'")
        found_ip = cache.get_ip(dep.vendor, dep.target_ip)
        if found_ip == None:
            common.fatal(f"Could not find IP dependency '{dep.vendor}/{dep.target_ip}'")
        else:
            check_dependencies(found_ip)


def cmp_dependencies(ip, sim_job):
    global bar
    global est_time
    global num_deps_to_compile
    deps = ip.get_ordered_deps()
    deps_to_cmp = []
    for dep in deps:
        common.dbg(f"Processing dep {dep.vendor}/{dep.name} to be compiled")
        if dep.name == "uvm":
            continue
        #if not dep.is_compiled[sim_job.simulator]:
        deps_to_cmp.append(dep)
    num_deps = len(deps_to_cmp)
    if num_deps > 0:
        if num_deps == 1:
            msg = f"Compiling 1 dependency"
        else:
            msg = f"Compiling {num_deps} dependencies"
        common.info(msg)
    else:
        return 0
    
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    defines = sim_job.cmp_args
    est_time = num_deps
    num_deps_to_compile = num_deps
    with tqdm(deps_to_cmp) as pbar:
        for dep in tqdm(deps_to_cmp):
            dep_str = f"{dep.vendor}/{dep.name}"
            pbar.set_description(dep_str)
            eal.compile_ip(dep, sim_job)
    #with alive_bar(num_deps, bar = 'smooth', stats="{eta} estimated", monitor=True, elapsed=True) as bar:
    #    for dep in deps_to_cmp:
    #        dep_str = f"{dep.vendor}/{dep.name}"
    #        bar.text(dep_str)
    #        eal.compile_ip(dep, sim_job)
    #        bar()
    return num_deps_to_compile


def cmp_dep(dep, sim_job):
    global num_deps_to_compile
    dep_str = f"{dep.vendor}/{dep.name}"
    common.dbg(f"Compiling IP dependency '{dep_str}'")
    eal.compile_ip(dep, sim_job)
    sem.acquire()
    bar.text(dep_str)
    num_deps_to_compile = num_deps_to_compile - 1
    sem.release()


def cmp_dut(ip, sim_job):
    defines = sim_job.cmp_args
    
    if ip.dut_ip_type == "fsoc":
        if not ip.dut_core.is_installed or sim_job.fsoc:
            flist_path = eal.invoke_fsoc(ip, ip.dut_core, sim_job)
            ip.dut_core.is_installed = True
        else:
            common.info("Skipping processing of DUT FuseSoC core '" + ip.dut_fsoc_full_name + "'.")
        eal.compile_fsoc_core(flist_path, ip.dut_core, sim_job)
    else:
        if ip.dut != None:
            if ip.dut.target_ip_model == None:
                common.fatal(f"Did not resolve DUT dependency ('{ip.dut.vendor}/{ip.dut.target_ip}')!")
            dut_ip_str = f"{ip.dut.vendor}/{ip.dut.target_ip}"
            compile_dut = False
            if ip.dut.target_ip_model.is_local:
                common.dbg("Found local IP DUT '" + dut_ip_str + "'")
                compile_dut = True
            else:
                common.dbg("Found external IP DUT '" + dut_ip_str + "'")
                if not ip.dut.target_ip_model.is_compiled[sim_job.simulator]:
                    compile_dut = True
            if compile_dut:
                if ip.dut.target_ip_model.sub_type == "vivado":
                    common.info("Compiling Vivado Project DUT IP '" + dut_ip_str + "'")
                    eal.compile_vivado_project(ip.dut.target_ip_model, sim_job)
                else:
                    common.info("Compiling DUT IP '" + dut_ip_str + "'")
                    eal.compile_ip(ip.dut.target_ip_model, sim_job)
            else:
                common.info("Skipping compilation of DUT IP '" + dut_ip_str + "'.")
        else:
            common.fatal("Could not find DUT IP '" + dut_ip_str + "'.")


def cmp_target_ip(ip, sim_job):
    defines = sim_job.cmp_args
    ip_str = f"{ip.vendor}/{ip.name}"
    if ip.sub_type == "vivado":
        #common.info(f"Compiling Vivado Project '{ip_str}'")
        eal.compile_vivado_project(ip_yml, sim_job)
    else:
        #common.info(f"Compiling '{ip_str}'")
        eal.compile_ip(ip, sim_job)



def print_end_of_compilation_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.elaborate:
        common.info("************************************************************************************************************************")
        common.info("* Compilation results: " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log")
    else:
        common.info("************************************************************************************************************************")
        common.info("* Compilation results:")
        common.info("************************************************************************************************************************")
        common.info("  emacs " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log &")
        common.info("  gvim  " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log &")
        common.info("  vim   " + cfg.sim_dir + "/cmp/" + ip_str + "." + sim_str + ".cmp.log")
        common.info("")


def print_end_of_elaboration_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    if sim_job.simulate:
        common.info("************************************************************************************************************************")
        common.info("* Elaboration results: " + cfg.sim_dir + "/elab/" + ip_str + "." + sim_str + ".elab.log")
    else:
        common.info("************************************************************************************************************************")
        common.info("* Elaboration results")
        common.info("************************************************************************************************************************")
        common.info("  emacs " + cfg.sim_dir + "/elab/" + ip_str + "." + sim_str + ".elab.log &")
        common.info("  gvim  " + cfg.sim_dir + "/elab/" + ip_str + "." + sim_str + ".elab.log &")
        common.info("  vim   " + cfg.sim_dir + "/elab/" + ip_str + "." + sim_str + ".elab.log")
        common.info("")


def print_end_of_simulation_message(ip, sim_job):
    if sim_job.dry_run:
        return
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    results_path = sim_job.results_path
    common.info("************************************************************************************************************************")
    common.info("* Simulation results")
    common.info("************************************************************************************************************************")
    if (sim_job.waves):
        if sim_job.simulator == common.simulators_enum.VIVADO:
            common.info("* Waveforms: $MIO_VIVADO_HOME/xsim -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.VCS:
            common.info("* Waveforms: $MIO_VCS_HOME/dve -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.METRICS:
            common.info(f"* Waveforms: $MIO_METRICS_HOME/mdc view wave {sim_job.results_dir_name}.vcd &")
        elif sim_job.simulator == common.simulators_enum.XCELIUM:
            common.info("* Waveforms: $MIO_XCELIUM_HOME/simvision -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.QUESTA:
            common.info("* Waveforms: $MIO_QUESTA_HOME/visualizer -gui " + results_path + "/waves.wdb &")
        elif sim_job.simulator == common.simulators_enum.RIVIERA:
            common.info("* Waveforms: $MIO_RIVIERA_HOME/??? -gui " + results_path + "/waves.wdb &")
        common.info("")
    common.info("* Main log: emacs " + results_path + "/sim.log &")
    common.info("            gvim  " + results_path + "/sim.log &")
    common.info("            vim   " + results_path + "/sim.log")
    common.info("")
    common.info("* Transaction logs: pushd " + results_path + "/trn_log")
    common.info("* Test result dir : pushd " + results_path)
    common.info("")





#!/usr/bin/python3
########################################################################################################################
# Copyright 2021-2023 Datum Technology Corporation
# All rights reserved
########################################################################################################################


########################################################################################################################
# IMPORTS
########################################################################################################################
from datetime import date
import os
import sys
import re
import jinja2
from jinja2 import Template
from mio import common
from mio import new
from mio import cfg
#from vsdx import VisioFile




########################################################################################################################
# GLOBALS
########################################################################################################################
dbg = False
uvm_gen_dir = re.sub("new_block.py", "", os.path.realpath(__file__)) + ".."
relative_path_to_template = uvm_gen_dir + "/templates/sets/"
out_path = ""
default_license = "All rights reserved."
name_of_copyright_owner = ""
name = ""
name_normal_case = ""


########################################################################################################################
# TEMPLATE DATA
########################################################################################################################
agent_cp_parameters = { }
agent_dp_in_parameters = { }
agent_dp_out_parameters = { }
env_tb_parameters = { }


########################################################################################################################
# MAIN
########################################################################################################################
def pick_out_path():
    global out_path
    global default_copyright_owner
    if len(sys.argv) > 1:
        out_path = sys.argv[1]
        print("Code will be output to " + out_path)
    if len(sys.argv) > 2:
        default_copyright_owner = sys.argv[2].replace('"', "")
        print("Default copyright owner is " + default_copyright_owner)


def prompt_user_values():
    global out_path
    global name
    global name_normal_case
    global name_of_copyright_owner
    global default_copyright_owner
    global agent_cp_parameters
    global agent_dp_in_parameters
    global agent_dp_out_parameters
    global env_tb_parameters
    
    default_copyright_owner = cfg.org_full_name
    default_vendor          = cfg.org_name

    common.info("Moore.io Template Generator: RTL Block UVM DV Combo (v1p0)")
    common.info("**********************************************************")
    common.info("The answers to the following questionnaire will be used to generate the code for your new UVM Block-Level Env+TB")
    common.info("")
    
    if out_path == "":
        out_path = common.prompt("Please enter the destination path (default: '.'):\n").strip()
        if out_path == "":
            out_path = "."
    
    name_of_copyright_owner = common.prompt("Please enter the name of the copyright holder or hit RETURN for the default (default is '" + default_copyright_owner + "'):\n").strip()
    if name_of_copyright_owner == "":
        name_of_copyright_owner = default_copyright_owner
    
    name_of_vendor = common.prompt("Please enter a vendor name or hit RETURN for the default (default is '" + default_vendor + "'):\n").strip()
    if name_of_vendor == "":
        name_of_vendor = default_vendor
    
    license = common.prompt("Please enter a usage license or hit RETURN for the default (default is '" + default_license + "'):\n").strip()
    if license == "":
        license = default_license
    
    name = common.prompt("Please enter the name for this block (ex: 'abc'); this name will be used for all UVM types (ex: 'uvme_abc_env_c'):\n").lower().strip()
    if name == "":
        sys.exit("ERROR: package name cannot be empty.  Exiting.")
    
    name_normal_case = common.prompt("Please enter the (descriptive) name for this block (ex: 'Advanced Bus Control'):\n").strip()
    if name_normal_case == "":
        sys.exit("ERROR: descriptive name cannot be empty.  Exiting.")
    
    agent_cp_parameters = {
        "name"                    : name + "_cp",
        "name_uppercase"          : name.upper() + "_CP",
        "name_normal_case"        : name_normal_case + " Block Control Plane",
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license,
        "year"                    : date.today().year,
        "vendor"                  : name_of_vendor
    }
    
    agent_dp_in_parameters = {
        "name"                    : name + "_dpi",
        "name_uppercase"          : name.upper() + "_DPI",
        "name_normal_case"        : name_normal_case + " Block Data Plane Input",
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license,
        "year"                    : date.today().year,
        "vendor"                  : name_of_vendor
    }
    
    agent_dp_out_parameters = {
        "name"                    : name + "_dpo",
        "name_uppercase"          : name.upper() + "_DPO",
        "name_normal_case"        : name_normal_case + " Block Data Plane Output",
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license,
        "year"                    : date.today().year,
        "vendor"                  : name_of_vendor
    }
    
    env_tb_parameters = {
        "name"                    : name,
        "name_uppercase"          : name.upper(),
        "name_normal_case"        : name_normal_case,
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license,
        "year"                    : date.today().year,
        "vendor"                  : name_of_vendor
    }


def gen_block():
    common.fatal("Not yet implemented")


def print_end_message():
    global out_path
    global name
    global parameters
    
    common.banner("IPs successfully generated:")
    common.info("  * " + out_path + "/uvma_" + name + "_cp")
    common.info("  * " + out_path + "/uvma_" + name + "_dpi")
    common.info("  * " + out_path + "/uvma_" + name + "_dpo")
    common.info("  * " + out_path + "/uvme_" + name)
    common.info("  * " + out_path + "/uvmt_" + name)


########################################################################################################################
# ENTRY POINT
########################################################################################################################
def main():
    #pick_out_path()
    prompt_user_values()
    gen_block()
    print_end_message()



# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import common
from mio import sim
from mio import eal
from mio import cache
from mio import cov
from mio import cfg
from mio import results
from mio import cli

import yaml
from yaml.loader import SafeLoader
from tqdm import tqdm
from multiprocessing import Process
import threading
from threading import BoundedSemaphore
from threading import Thread
import os
import random
import math
import time
from datetime import datetime


bar = None
sem = None
sem_cfg = None


class TestSuite:
    """TestSuite model"""

    def __init__(self, ip, filepath, simulator):
        self.timestamp         = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        self.ip                = ip
        self.filepath          = filepath
        self.name              = ""
        self.ip_name           = ""
        self.cmp_args          = []
        self.elab_args         = []
        self.sim_args          = []
        self.verbosity         = []
        self.sets              = {}
        self.regr              = {}
        self.waves             = []
        self.cov               = []
        self.max_durations     = []
        self.max_jobs          = []
        if simulator == "viv":
            self.simulator = common.simulators_enum.VIVADO
        elif simulator == "mdc":
            self.simulator = common.simulators_enum.METRICS
        elif simulator == "vcs":
            self.simulator = common.simulators_enum.VCS
        elif simulator == "xcl":
            self.simulator = common.simulators_enum.XCELIUM
        elif simulator == "qst":
            self.simulator = common.simulators_enum.QUESTA
        elif simulator == "riv":
            self.simulator = common.simulators_enum.RIVIERA
        else:
            self.simulator = cfg.default_simulator
    
    def parse_yml(self, yml):
        if 'test-suite' not in yml:
            common.fatal("No 'test-suite' entry!")
        common.dbg("Parsing test-suite metadata")
        
        if 'info' not in yml['test-suite']:
            common.fatal("No 'info' entry in 'test-suite'!")
        common.dbg("Parsing test-suite info")
        
        if 'name' not in yml['test-suite']['info']:
            common.fatal("No 'name' entry in 'test-suite'!")
        self.name = yml['test-suite']['info']['name'].strip()
        if not isinstance(self.name, str):
            common.fatal("'name' must be a string!")
        
        if 'ip' not in yml['test-suite']['info']:
            common.fatal("No 'ip' entry in 'test-suite'!")
        self.ip_name = yml['test-suite']['info']['ip'].strip().lower()
        if not isinstance(self.ip_name, str):
            common.fatal("'ip_name' must be a string!")
        
        if 'target' in yml['test-suite']:
            common.dbg("Parsing test-suite target")
            if 'cmp-args' in yml['test-suite']['target']:
                self.cmp_args = yml['test-suite']['target']['cmp-args']
                if type(self.cmp_args) is not list and self.cmp_args != None:
                    common.fatal("'cmp-args' must be a list!")
                for arg in self.cmp_args:
                    if not isinstance(arg, str):
                        common.fatal("'cmp-args' entry key is not a string")
                    if not isinstance(self.cmp_args[arg], str):
                        common.fatal("'cmp-args' entry is not a string")
            
            if 'elab-args' in yml['test-suite']['target']:
                self.elab_args = yml['test-suite']['target']['elab-args']
                if type(self.elab_args) is not list and self.elab_args != None:
                    common.fatal("'elab-args' must be a list!")
                for arg in self.elab_args:
                    if not isinstance(arg, str):
                        common.fatal("'elab-args' entry key is not a string")
                    if not isinstance(self.elab_args[arg], str):
                        common.fatal("'elab-args' entry is not a string")
            
            if 'sim-args' in yml['test-suite']['target']:
                self.sim_args = yml['test-suite']['target']['sim-args']
                if type(self.sim_args) is not list and self.sim_args != None:
                    common.fatal("'sim-args' must be a list!")
                for arg in self.sim_args:
                    if not isinstance(arg, str):
                        common.fatal("'sim-args' entry key is not a string")
                    if not isinstance(self.sim_args[arg], str):
                        common.fatal("'sim-args' entry is not a string")
        
        if 'settings' not in yml['test-suite']:
            common.fatal("No 'settings' entry in 'test-suite'!")
        common.dbg("Parsing test-suite settings")
        
        if 'waves' in yml['test-suite']['settings']:
            self.waves = yml['test-suite']['settings']['waves']
            if type(self.waves) is not list:
                common.fatal("'waves' must be a list!")
                for wave in self.waves:
                    if not isinstance(wave, str):
                        common.fatal("'waves' entry is not a string")
        
        if 'cov' in yml['test-suite']['settings']:
            self.cov = yml['test-suite']['settings']['cov']
            if type(self.cov) is not list:
                common.fatal("'cov' must be a list!")
                for cov in self.cov:
                    if not isinstance(cov, str):
                        common.fatal("'cov' entry is not a string")
        
        if 'verbosity' in yml['test-suite']['settings']:
            self.verbosity = yml['test-suite']['settings']['verbosity']
            if type(self.verbosity) is not dict:
                common.fatal("'max-duration' must be a dictionary!")
            for verb in self.verbosity:
                if not isinstance(verb, str):
                    common.fatal("'verbosity' entry key is not a string!")
                if not (self.verbosity[verb] in cli.uvm_levels):
                    common.fatal(f"'verbosity' entry '{self.verbosity[verb]}' is not valid! Choices are: none, low, medium, high or debug.")
        
        if 'max-duration' in yml['test-suite']['settings']:
            self.max_durations = yml['test-suite']['settings']['max-duration']
            if type(self.max_durations) is not dict:
                common.fatal("'max-duration' must be a dictionary!")
            for duration in self.max_durations:
                if not isinstance(duration, str):
                    common.fatal("'max-duration' entry key is not a string!")
                if not (type(self.max_durations[duration]) == int or float):
                    common.fatal("'max-duration' entry is not a number!")
        else:
            common.fatal("No 'max-duration' entry in test-suite settings!")
        
        if 'max-jobs' in yml['test-suite']['settings']:
            self.max_jobs = yml['test-suite']['settings']['max-jobs']
            if type(self.max_jobs) is not dict:
                common.fatal("'max-jobs' must be a dictionary!")
            for max in self.max_jobs:
                if not isinstance(max, str):
                    common.fatal("'max-jobs' entry key is not a string!")
                if not (type(self.max_jobs[max]) == int):
                    common.fatal("'max-jobs' entry is not an integer!")
        else:
            common.fatal("No 'max-jobs' entry in test-suite settings!")
        
        set_count = 0
        common.dbg("Parsing sets")
        for set in yml:
            if set == "test-suite":
                continue
            if type(yml[set]) is not dict:
                common.fatal(f"Test set '{set}' is not a dictionary!")
            test_set = TestSet(set, self)
            self.add_test_set(test_set)
            common.dbg(f"Parsing set '{set}'")
            for group in yml[set]:
                if type(yml[set][group]) is not dict:
                    common.fatal(f"Test group '{set}.{group}' is not a dictionary!")
                common.dbg(f"Parsing group '{set}.{group}'")
                group_obj = TestGroup(group, test_set)
                test_set.add_test_group(group_obj)
                for test in yml[set][group]:
                    if type(yml[set][group][test]) is not dict:
                        common.fatal(f"Test '{set}.{group}.{test}' is not a dictionary!")
                    for regr in yml[set][group][test]:
                        regr_item = yml[set][group][test][regr]
                        
                        if regr not in self.regr:
                            regression = Regression(regr.strip().lower(), group_obj, set, self)
                            self.add_regression(regression)
                            common.dbg(f"Added regression '{regr}' from '{set}.{group}.{test}'")
                        else:
                            regression = self.regr[regr]
                        args = []
                        
                        if type(regr_item) is not dict:
                            # Inline ("simple") regression entry
                            if type(regr_item) is int:
                                # Specifying how many seeds to run
                                if regr_item <= 0:
                                    common.fatal(f"value of '{regr_item}' is less than 1: '{set}.{group}.{test}.{regr}'")
                                for ii in range(regr_item):
                                    seed = random.randint(1, 2147483646)
                                    regression.add_test(test, seed, args)
                            elif type(regr_item) is list:
                                # Specifying specific seeds to run
                                for seed in regr_item:
                                    if not (type(seed) == int):
                                        common.fatal(f"seed value '{str(seed)}' is not an integer: '{set}.{group}.{test}.{regr}'!")
                                    regression.add_test(test, seed, args)
                            else:
                                common.fatal(f"Illegal regression entry: '{set}.{group}.{test}.{regr}'")
                        else:
                            # Complex regression spec
                            if 'seeds' not in regr_item:
                                common.fatal(f"Regression entry missing 'seeds' entry: '{set}.{group}.{test}.{regr}'")
                            else:
                                if 'args' in regr_item:
                                    # Sim arguments are specified
                                    if type(regr_item['args']) is not list:
                                        common.fatal(f"Regression entry args is not list: '{set}.{group}.{test}.{regr}'")
                                    args = regr_item['args']
                                    for arg in args:
                                        if not isinstance(args[arg], str):
                                            common.fatal(f"Regression entry args entry is not string: '{set}.{group}.{test}.{regr}'")
                                
                                if type(regr_item['seeds']) is int:
                                    # Specifying how many seeds to run
                                    if regr_item['seeds'] <= 0:
                                        common.fatal(f"'seeds' value '{regr_item['seeds']}' is less than 1: '{set}.{group}.{test}.{regr}'")
                                    for ii in range(regr_item['seeds']):
                                        seed = random.randint(1, 2147483646)
                                        regression.add_test(test, seed, args)
                                elif type(regr_item['seeds']) is list:
                                    # Specifying specific seeds to run
                                    for seed in regr['seeds']:
                                        if not (type(seed) == int):
                                            common.fatal(f"seed value '{str(seed)}' is not an integer: '{set}.{group}.{test}.{regr}'!")
                                        regression.add_test(test, seed, args)
                                else:
                                    common.fatal(f"Illegal regression entry: '{set}.{group}.{test}.{regr}'")

    def check_consistency(self):
        if self.ip_name != self.ip.name:
            common.fatal(f"IP '{self.ip_name}' in suite does not match '{self.ip.name}' as specified!")
        if len(self.sets) == 0:
            common.fatal("Test suite does not contain any test sets!")
        test_count = 0
        for regr in self.regr:
            if (self.regr[regr].name == cfg.regression_name):
                test_count = len(self.regr[regr].tests)
        if test_count == 0:
            common.fatal("Regression describes no tests.  Please add entries and try again.")
        if test_count > 1000:
            common.fatal("Regression describes more than 1,000 tests.  Please reduce this and try again.")
        if test_count > 100:
            common.warning(f"Regression describes more than 100 tests ({test_count})")
        #for wave in self.waves:
        #    if wave not in self.regr:
        #        common.fatal(f"'waves' entry '{wave}' is not a regression declared in this test suite")
        #for cov in self.cov:
        #    if cov not in self.regr:
        #        common.fatal(f"'cov' entry '{cov}' is not a regression declared in this test suite")
        #for verb in self.verbosity:
        #    if verb not in self.regr:
        #        common.fatal(f"'verbosity' entry '{verb}' is not a regression declared in this test suite")
        for max in self.max_durations:
            #if max not in self.regr:
            #    common.fatal(f"'max-duration' entry '{max}' is not a regression declared in this test suite")
            if max in self.regr:
                self.regr[max].max_duration = self.max_durations[max]
        for max in self.max_jobs:
            #if max not in self.regr:
            #    common.fatal(f"'max-jobs' entry '{max}' is not a regression declared in this test suite")
            if max in self.regr:
                self.regr[max].max_jobs = self.max_jobs[max]
    
    def add_test_set(self, set):
        if set.name in self.sets:
            common.fatal(f"Test group '{group.name}' already exists within test set '{self.name}'")
        else:
            self.sets[set.name] = set
            set.suite = self
    
    def add_regression(self, regr):
        self.regr[regr.name] = regr
        regr.suite = self
        common.dbg(f"Regression '{regr.name}' added to set '{self.name}'")
    
    def get_cmp_job(self):
        sim_job = sim.SimulationJob(f"{self.ip.vendor}/{self.ip.name}")
        sim_job.is_regression   = True
        sim_job.regression_name = cfg.regression_name
        sim_job.regression_timestamp = self.timestamp
        sim_job.one_shot  = False
        sim_job.compile   = True
        sim_job.elaborate = False
        sim_job.simulate  = False
        sim_job.gui       = False
        sim_job.simulator = self.simulator
        sim_job.raw_args  = self.cmp_args
        if cfg.regression_name in self.waves:
            sim_job.waves = True
        else:
            sim_job.waves = False
        if cfg.regression_name in self.cov:
            sim_job.cov = True
        else:
            sim_job.cov = False
        return sim_job
    
    def get_elab_job(self):
        sim_job = sim.SimulationJob(f"{self.ip.vendor}/{self.ip.name}")
        sim_job.is_regression   = True
        sim_job.regression_name = cfg.regression_name
        sim_job.regression_timestamp = self.timestamp
        sim_job.one_shot  = False
        sim_job.compile   = False
        sim_job.elaborate = True
        sim_job.simulate  = False
        sim_job.gui       = False
        sim_job.simulator = self.simulator
        sim_job.raw_args  = self.elab_args
        if cfg.regression_name in self.waves:
            sim_job.waves = True
        else:
            sim_job.waves = False
        if cfg.regression_name in self.cov:
            sim_job.cov = True
        else:
            sim_job.cov = False
        return sim_job
    
    def get_regression(self, regr_name):
        for reg in self.regr:
            if self.regr[reg].name == regr_name:
                return self.regr[reg]
        commmon.fatal(f"Could not find regression '{regr_name}'")


class TestSet:
    """Test Set model"""

    def __init__(self, name, suite):
        self.name         = name
        self.suite        = suite
        self.waves        = False
        self.cov          = False
        self.max_duration = 1
        self.max_jobs     = 1
        self.groups       = {}
    
    def add_test_group(self, group):
        if group.name in self.groups:
            common.fatal(f"Test group '{group.name}' already exists within test set '{self.name}'")
        else:
            self.groups[group.name] = group
            group.set = self
            common.dbg(f"Test group '{group.name}' added to set '{self.name}'")


class TestGroup:
    """Test Group model"""

    def __init__(self, name, set):
        self.name  = name
        self.set   = set
        self.tests = []
    
    def add_test(self, test):
        self.tests.append(test)
        common.fatal(f"Test '{test.name}' added to group '{self.name}'")


class Regression:
    """Regression model"""

    def __init__(self, name, group, set, suite):
        self.name         = name
        self.tests        = []
        self.set          = set
        self.group        = group
        self.suite        = suite
        self.max_duration = 0
        self.max_jobs     = 1
    
    def add_test(self, name, seed, args):
        test = RegressionTest(name, self, self.group, self.set, self.suite, seed, args)
        self.tests.append(test)
        common.dbg(f"Test '{name}' added to group '{self.group.name}'")
    
    def get_tests(self):
        return self.tests
    
    def reduce(self):
        removal_list = []
        ii = 0
        for curr_test in self.tests:
            jj = 0
            for other_test in self.tests:
                if other_test == curr_test:
                    continue
                else:
                    if (jj not in removal_list):
                        if curr_test.is_equal(other_test):
                            removal_list.append(ii)
                jj += 1
            ii += 1
        if len(removal_list) > 0:
            common.warning(f"Found {len(removal_list)} redundancies in regression '{name}'")
            for index in removal_list:
                self.tests[index].remove(index)


class RegressionTest:
    """Regression Test model"""

    def __init__(self, name, regression, test_group, set, suite, seed, args):
        self.name       = name
        self.regression = regression
        self.test_group = test_group
        self.set        = set
        self.suite      = suite
        self.seed       = seed
        self.args       = args
    
    def get_sim_job(self):
        sim_job = sim.SimulationJob(f"{self.suite.ip.vendor}/{self.suite.ip.name}")
        sim_job.is_regression   = True
        sim_job.regression_name = cfg.regression_name
        sim_job.regression_timestamp = self.suite.timestamp
        sim_job.test      = self.name
        sim_job.seed      = self.seed
        sim_job.compile   = False
        sim_job.elaborate = False
        sim_job.simulate  = True
        sim_job.gui       = False
        sim_job.simulator = self.suite.simulator
        sim_job.raw_args  = self.args
        if cfg.regression_name in self.suite.waves:
            sim_job.waves = True
        else:
            sim_job.waves = False
        if cfg.regression_name in self.suite.cov:
            sim_job.cov = True
        else:
            sim_job.cov = False
        if cfg.regression_name in self.suite.verbosity:
            sim_job.verbosity = self.suite.verbosity[cfg.regression_name]
        else:
            sim_job.verbosity = "medium"
        
        return sim_job
    
    def is_equal(self, other):
        if self.name != other.name:
            return False
        if self.regression != other.regression:
            return False
        if self.test_group != other.test_group:
            return False
        if self.set != other.set:
            return False
        if self.suite != other.suite:
            return False
        if self.seed != other.seed:
            return False
        if self.args.sort() != other.args.sort():
            return False
        return True


def main(ip_str, regression, simulator, dry_mode):
    vendor, name = common.parse_dep(ip_str)
    if vendor == "":
        ip = cache.get_anon_ip(name, True)
    else:
        ip = cache.get_ip(vendor, name, True)
    deps_to_install = len(ip.get_deps_to_install())
    if deps_to_install > 0:
        common.fatal(f"You must first install this IP's dependencies ({deps_to_install}): `mio install {name}`")
    
    timestamp_start = datetime.now()
    test_suite = scan_target_ip_for_test_suite(ip, simulator)
    regression = test_suite.get_regression(cfg.cli_args.regr)
    regression.reduce()
    tests      = regression.get_tests()
    prep_target_ip(ip, test_suite)
    
    sim_job_list = []
    for test in tests:
        sim_job_list.append(test.get_sim_job())
    
    launch_sim_jobs(ip, test_suite, regression, sim_job_list, dry_mode)
    timestamp_end = datetime.now()
    if not dry_mode:
        regr_results = results.main(f"{ip.vendor}/{ip.name}", "", True, test_suite.name, regression.name, test_suite.timestamp)
        cov_report_path = cov.gen_cov_report(f"{ip.vendor}/{ip.name}", True, test_suite.name, regression.name, test_suite.timestamp)
        print_end_of_regression_msg(ip, regr_results, cov_report_path, test_suite, regression, sim_job_list, timestamp_start, timestamp_end)


def launch_sim_jobs(ip, test_suite, regression, sim_job_list, dry_mode):
    global bar
    global sem
    global sem_cfg
    threads = []
    regression_name = ""
    if cfg.test_suite_name == "":
        regression_name = f"{cfg.regression_name}"
        common.banner(f"Running regression '{cfg.regression_name}': {str(len(sim_job_list))} test(s) with {str(regression.max_duration)} hour(s) timeout")
    else:
        regression_name = f"{cfg.test_suite_name}.{cfg.regression_name}"
        common.banner(f"Running regression '{cfg.regression_name}' from test suite '{cfg.test_suite_name}': {str(len(sim_job_list))} test(s) with {str(regression.max_duration)} hour(s) timeout")
    
    common.create_dir(cfg.regr_results_dir + "/" + cfg.target_ip_name + "_" + regression_name)
    common.create_dir(cfg.regr_results_dir + "/" + cfg.target_ip_name + "_" + regression_name + "/" + test_suite.timestamp)
    sem = BoundedSemaphore(regression.max_jobs)
    sem_cfg = BoundedSemaphore(1)
    with tqdm(sim_job_list) as bar:
    #with alive_bar(len(sim_job_list), bar = 'smooth') as bar:
        for sim_job in sim_job_list:
            common.dbg("Launching test threads")
            thread = Thread(target=launch_test, args=(ip,test_suite,sim_job,dry_mode,))
            thread.daemon = True
            threads.append(thread)
            thread.start()
            time.sleep(10) # HACK!!!!! or else threads step on top of each other
        timeout = Thread(target=timeout_process, args=(regression.max_duration,))
        timeout.daemon = True
        timeout.start()
        common.dbg("Waiting for threads to end")
        for thread in threads:
            if thread.is_alive():
                thread.join()


def timeout_process(hours):
    total_minutes = 0
    common.dbg(f"Starting timeout process for {str(hours)} hour(s)")
    for n in range(math.ceil(hours*60)):
        time.sleep(60)
        total_minutes += 1
    common.fatal(f"Regression timed out after {str(hours)} hour(s)")


def prep_target_ip(ip, test_suite):
    if cfg.test_suite_name != "":
        regression_name = cfg.test_suite_name + "_" + cfg.regression_name
    else:
        regression_name = cfg.regression_name
    cmp_job  = test_suite.get_cmp_job ()
    elab_job = test_suite.get_elab_job()
    cmp_job .regression_name = regression_name
    elab_job.regression_name = regression_name
    sim.main(cmp_job )
    sim.main(elab_job)


def launch_test(ip, test_suite, sim_job, dry_mode):
    global sem_cfg
    sem_cfg.acquire()
    if cfg.test_suite_name != "":
        regression_name = cfg.test_suite_name + "_" + cfg.regression_name
    else:
        regression_name = cfg.regression_name
    sim_job.regression_name = regression_name
    plus_args = sim.convert_cli_args_to_plusargs(sim_job)
    common.dbg("Waiting turn for simulating:\n" + str(sim_job))
    sem_cfg.release()
    wait_for_turn()
    common.dbg("Done waiting. Starting thread for simulating:\n" + str(sim_job))
    if dry_mode:
        common.info(f"  dry-run: test='{sim_job.test}' seed='{str(sim_job.seed)}' args='{str(sim_job.args)}' waves='{str(sim_job.waves)}' cov='{str(sim_job.cov)}'")
    else:
        eal.simulate(ip, sim_job)
    common.dbg("Done simulating:\n" + str(sim_job))
    bar.update(1)
    done_with_turn()


def wait_for_turn():
    global sem
    sem.acquire()


def done_with_turn():
    global sem
    sem.release()



def scan_target_ip_for_test_suite(ip, simulator):
    ip_str = f"{ip.vendor}/{ip.name}"
    cfg.regression_name = cfg.cli_args.regr
    suite = None
    test_suite_name = ""
    if "." in cfg.regression_name:
        split_string = regr_name.split(".")
        cfg.test_suite_name = split_string[0]
        cfg.regression_name = split_string[1]
        common.dbg(f"Test suite name is '{test_suite_name}' and regression name is '{cfg.regression_name}'")
    else:
        common.dbg(f"Test suite name is empty and regression name is '{cfg.regression_name}'")
    
    if ip.hdl_src_tests_path == "":
        common.fatal(f"IP '{ip_str}' does not have 'hdl-src.tests-path' defined in its ip.yml")
    ip_tests_path = ip.path + "/" + ip.src_path + "/" + ip.hdl_src_tests_path
    for dirpath, dirnames, filenames in os.walk(ip_tests_path):
        for file in filenames:
            if cfg.test_suite_name == "":
                common.dbg(f"looking at file {file}")
                if file.endswith("ts.yml"):
                    filepath = dirpath + "/" + file
                    common.dbg(f"Found target test suite descriptor file at {filepath}")
                    suite = parse_test_suite_file(ip, simulator, filepath)
                    break
            else:
                if file.endswith(".ts.yml"):
                    filepath = dirpath + "/" + file
                    ts_name = file.replace(".ts.yml", "")
                    ts_name = ts_name.replce(dirpath, "")
                    common.dbg(f"Found test suite '{ts_name}' descriptor file at {filepath}")
                    if ts_name == cfg.test_suite_name:
                        common.dbg(f"Found target test suite '{ts_name}' descriptor file at {filepath}")
                        suite = parse_test_suite_file(ip, simulator, filepath)
                        break
    if suite == None:
        common.fatal("Could not find test suite")
    return suite


def parse_test_suite_file(ip, simulator, file):
    test_suite = None
    try:
        with open(file, 'r') as yamlfile:
            ts_yaml = yaml.load(yamlfile, Loader=SafeLoader)
            if ts_yaml:
                test_suite = TestSuite(ip, file, simulator)
                test_suite.parse_yml(ts_yaml)
                test_suite.check_consistency()
            else:
                common.fatal(f"Failed to parse test suite yml '{file}':" + str(e))
    except Exception as e:
        common.fatal(f"Failed to parse test suite '{file}': " + str(e))
    return test_suite


def print_end_of_regression_msg(ip, results, cov_report_path, test_suite, regression, sim_job_list, timestamp_start, timestamp_end):
    duration = timestamp_end - timestamp_start
    hours = math.floor(duration.total_seconds() / 3600)
    minutes = math.floor((duration.total_seconds() - (hours*3600)) / 60)
    seconds = math.ceil(duration.total_seconds() - (hours*3600) - (minutes*60))
    duration_str = f"{str(hours)} hour(s), {str(minutes)} minute(s), {str(seconds)} second(s)"
    
    if cfg.test_suite_name != "":
        regression_name = cfg.test_suite_name + "_" + cfg.regression_name
    else:
        regression_name = cfg.regression_name
    
    results_path = cfg.regr_results_dir + "/" + ip.name + "_" + regression.name + "/" + test_suite.timestamp
    
    if results.passed:
        common.info("************************************************************************************************************************")
        common.info(f"  \033[32m\033[1m'{regression_name}' regression PASSED\033[0m")
        common.info("************************************************************************************************************************")
        common.info(f"  Duration: {duration_str}")
    else:
        common.info("************************************************************************************************************************")
        common.info(f"  \033[31m\033[1m'{regression_name}' regression FAILED: {str(results.num_failed_tests)} failure(s)\033[0m")
        common.info("************************************************************************************************************************")
        common.info(f"  Duration      : {duration_str}")
        common.info(f"  # tests passed: {str(results.num_passed_tests)}")
        common.info(f"  Passing rate  : {str(results.pct_passed)} %")
    
    common.info(f"")
    common.info(f"  HTML report      : firefox {results.html_report_path} &")
    common.info(f"  Coverage report  : pushd   {cov_report_path}")
    common.info(f"  Results directory: pushd   {results_path}")




# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


version = "1.0.7"
version_text = f"Moore.io CLI Client v{version}"




main_help_text = f"""
                                 Moore.io (`mio`) Command Line Interface (CLI) - v{version}
                                      User Manual: https://mio-cli.readthedocs.io/
              https://mooreio.com - Copyright 2021-2023 Datum Technology Corporation - https://datumtc.ca
Usage:
  mio [--version] [--help]
  mio [--wd WD] [--dbg] CMD [OPTIONS]

Options:
  -v, --version
    Prints the mio version and exits.
  
  -h, --help
    Prints the overall synopsis and a list of the most commonly used commands and exits.
    
  -C WD, --wd WD
    Run as if mio was started in WD (Working Directory) instead of the Present Working Directory `pwd`.
   
  --dbg
    Enables debugging outputs from mio.

Full Command List (`mio help CMD` for help on a specific command):
   Help and Shell/Editor Integration
      doctor         Runs a set of checks to ensure mio installation has what it needs to operate properly
      help           Prints documentation for mio commands
   
   Project and Code Management
      init           Starts project creation dialog
      new            Creates new source code via the mio template engine
   
   IP and Credentials Management
      install        Install all IP dependencies from IP Marketplace
      login          Start session with IP Marketplace
      package        Create a compressed (and potentially encrypted) archive of an IP
      publish        Publish IP to IP Marketplace (must have mio admin account)
   
   EDA Automation
      !              Repeat last command
      regr           Runs regression against an IP
      sim            Performs necessary steps to simulate an IP with any simulator
      
   Manage Results and other EDA Tool Outputs
      clean          Manages outputs from tools (other than job results)
      cov            Manages coverage data from EDA tools
      dox            HDL source code documentation generation via Doxygen
      results        Manages results from EDA tools
"""




repeat_help_text = """Moore.io Repeat (!) Command
   Repeats last command ran by mio.  Currently only supports the `sim` command.
   
Usage:
   mio ! CMD [OPTIONS]
   
Options:
   -b, --bwrap  Does not run command, only creates shell script to re-create the command without mio and creates a
                tarball of the project outside the project root directory.  Currently only supports `sim` command.
   
Examples:
   mio sim uvmt_example -t rand_stim -s 1 ; mio ! sim -b  # Run a simulation for `uvmt_example` and create a tarball
                                                          # that can be run by anyone using only bash."""




sim_help_text = """Moore.io Sim(ulation) Command
   Performs necessary steps to run simulation of an IP.  Only supports Digital Logic Simulation for the time being.
   
   While the controls for individual steps (FuseSoC processing, compilation, elaboration and simulation) are exposed, it
   is recommended to let `mio sim` manage this process as much as possible.  In the event of corrupt simulator
   artifacts, see `mio clean`.  Combining any of the step-control arguments (-F, -C, -E, -S) with missing steps can
   result in unpredictable behavior and is not recommended (ex: `-FS` is illegal).
   
   Two types of arguments (--args) can be passed: compilation (+define+NAME[=VALUE]) and simulation (+NAME[=VALUE]).
   
   For running multiple tests in parallel, see `mio regr`.
   
Usage:
   mio sim IP [OPTIONS] [--args ARG ...]
   
Options:
   -t TEST     , --test      TEST       Specify the UVM test to be run.
   -s SEED     , --seed      SEED       Positive Integer. Specify randomization seed  If none is provided, a random one will be picked.
   -v VERBOSITY, --verbosity VERBOSITY  Specifies UVM logging verbosity: none, low, medium, high, debug. [default: medium]
   -e ERRORS   , --errors    ERRORS     Specifies the number of errors at which compilation/elaboration/simulation is terminated.  [default: 10]
   -a APP      , --app       APP        Specifies simulator application to use: viv, mdc, vcs, xcl, qst, riv. [default: viv]
                                        WARNING: Only Vivado is currently supported. VCS and Metrics will be added in the
                                                 near future (2022/08/23).
   -w          , --waves                Enable wave capture to disk.
   -c          , --cov                  Enable code & functional coverage capture.
   -g          , --gui                  Invokes simulator in graphical or 'GUI' mode.
   
   -S   Simulate  target IP.
   -E   Elaborate target IP.
   -C   Compile   target IP.
   -F   Invoke FuseSoC to prepare core(s) for compilation.
   
Examples:
   mio sim uvmt_my_ip -t smoke -s 1 -w -c             # Compile, elaborate and simulate test 'uvmt_my_ip_smoke_test_c'
                                                      # for IP 'uvmt_my_ip' with seed '1' and waves & coverage capture enabled.
   mio sim uvmt_my_ip -t smoke -s 1 --args +NPKTS=10  # Compile, elaborate and simulate test 'uvmt_my_ip_smoke_test_c'
                                                      # for IP 'uvmt_my_ip' with seed '1' and a simulation argument.
   mio sim uvmt_my_ip -S -t smoke -s 42 -v high -g    # Only simulates test 'uvmt_my_ip_smoke_test_c' for IP 'uvmt_my_ip'
                                                      # with seed '42' and UVM_HIGH verbosity using the simulator in GUI mode.
   mio sim uvmt_my_ip -C                              # Only compile 'uvmt_my_ip'.
   mio sim uvmt_my_ip -E                              # Only elaborate 'uvmt_my_ip'.
   mio sim uvmt_my_ip -CE                             # Compile and elaborate 'uvmt_my_ip'."""




regr_help_text = """Moore.io Regr(ession) Command
   Runs a set of tests against a specific IP.  Regressions are described in Test Suite files (`[<target>.]ts.yml`).
   
Usage:
   mio regr IP [TARGET.]REGRESSION [OPTIONS]
   
Options:
   -d, --dry-run  Compiles, elaborates, but only prints the tests mio would normally run (does not actually run them).
   
Examples:
   mio regr uvmt_my_ip sanity            # Run sanity regression for IP 'uvm_my_ip', from test suite 'ts.yml'
   mio regr uvmt_my_ip apb_xc.sanity     # Run sanity regression for IP 'uvm_my_ip', from test suite 'apb_xc.ts.yml'
   mio regr uvmt_my_ip axi_xc.sanity -d  # Dry-run sanity regression for IP 'uvm_my_ip', from test suite 'axi_xc.ts.yml'"""




clean_help_text = """Moore.io Clean Command
   Deletes output artifacts from EDA tools.  Only simulation is currently supported.
   
Usage:
   mio clean IP [OPTIONS]
   
Options:
   -d, --deep  Also clean compiled IP dependencies.
   
Examples:
   mio clean uvmt_my_ip     # Delete compilation, elaboration and simulation binaries for IP 'uvmt_my_ip'
   mio clean uvmt_my_ip -d  # Delete compilation, elaboration and simulation binaries for IP 'uvmt_my_ip' and its dependencies"""




dox_help_text = """Moore.io Dox(ygen) Invokation Command
   Generates reference documentation from IP HDL source code.
   
Usage:
   mio dox IP [OPTIONS]
   
Options:
   
Examples:
   mio dox my_ip  # Generates HTML documentation for IP 'my_ip'"""




init_help_text = """Moore.io Init(ialization) Command
   Creates a new Project skeleton if not already within a Project.  If so, a new IP skeleton is created.
   This is the recommended method for importing code to the Moore.io ecosystem.
   
Usage:
   mio init [OPTIONS]

Options:
   
Examples:
   mio              init  # Create a new empty Project/IP in this location.
   mio -C ~/my_proj init  # Create a new empty Project at a specific location."""




install_help_text = """Moore.io IP Install Command
   Installs an IP and any IPs that it depends on from the Moore.io IP Marketplace (https://mooreio.com).  IPs can be
   installed either locally (PROJECT_ROOT/.mio/vendors) or globally (~/.mio/vendors).
   
Usage:
   mio install IP [OPTIONS]
   
Options:
   -g         , --global             # Installs IP dependencies for all user projects
   -u USERNAME, --username USERNAME  # Specifies Moore.io username (must be combined with -p)
   -p PASSWORD, --password PASSWORD  # Specifies Moore.io password (must be combined with -u)
   
Examples:
   mio install uvmt_my_ip                          # Install IP dependencies for 'uvmt_my_ip' locally.
   mio install uvmt_another_ip                     # Install IP dependencies for 'uvmt_another_ip' globally.
   mio install uvmt_my_ip -u jenkins -p )Kq3)fkqm  # Specify credentials for Jenkins job."""




login_help_text = """Moore.io User Login Command
   Authenticates session with the Moore.io IP Marketplace (https://mooreio.com).
   
Usage:
   mio login [OPTIONS]
   
Options:
   -u USERNAME, --username USERNAME  # Specifies Moore.io username (must be combined with -p)
   -p PASSWORD, --password PASSWORD  # Specifies Moore.io password (must be combined with -u)
   
Examples:
   mio login                          # Asks credentials only if expired (or never entered)
   mio login -u jenkins -p )Kq3)fkqm  # Specify credentials inline"""




publish_help_text = """Moore.io IP Publish Command
   Packages and publishes an IP to the Moore.io IP Marketplace (https://mooreio.com).
   Currently only available to administrator accounts.
   
Usage:
   mio publish IP [OPTIONS]
   
Options:
   -u USERNAME, --username USERNAME  # Specifies Moore.io username (must be combined with -p)
   -p PASSWORD, --password PASSWORD  # Specifies Moore.io password (must be combined with -u)
   -o ORG     , --org      ORG       # Specifies Moore.io IP Marketplace Organization client name.  Commercial IPs only.
   
Examples:
   mio publish uvma_my_ip                               # Publish IP 'uvma_my_ip'.
   mio publish uvma_my_ip -u acme_jenkins -p )Kq3)fkqm  # Specify credentials for Jenkins job.
   mio publish uvma_my_ip -o chip_inc                   # Publish IP 'uvma_my_ip' for client 'chip_inc'."""




package_help_text = """Moore.io IP Package Command
   Command for encrypting/compressing entire IP on local disk.  To enable IP encryption, add an 'encrypted' entry to the
   'hdl-src' section of your descriptor (ip.yml).  Moore.io will only attempt to encrypt using the simulators listed
   under 'simulators-supported' of the 'ip' section.
   
   Vivado requires a key for encryption; please ensure that you have specified your key location either in the project
   or user Configuration file (mio.toml).  https://mooreio-client.readthedocs.io/en/latest/configuration.html#encryption
   for more on the subject.
   
Usage:
   mio package IP DEST [OPTIONS]
   
Options:
   -n, --no-tgz  # Do not create compressed tarball
   
Examples:
   mio package uvma_my_ip ~        # Create compressed archive of IP 'uvma_my_ip' under user's home directory.
   mio package uvma_my_ip ~/ip -n  # Process IP 'uvma_my_ip' but do not create compressed archive."""




new_help_text = """Moore.io New Command
   Invokes the Datum UVM Code Template system.  If no generator name is specified, the user is prompted to select from a
   list of what is currently available: 17 Templates, from advanced agents to tests.
    
Usage:
   mio new [OPTIONS]
   
Options:
   -t TEMPLATE, --template TEMPLATE  Name of template to use: basic_agent, parallel_agent, serial_agent, block_tb, ss_tb,
                                                              lib, ral, comp, obj, reg_adapter, reg_block, reg, seq_lib,
                                                              seq, test, vseq_lib, vseq
Examples:
   mio new              # Invoke the template menu
   mio new -t block_tb  # Create new UVM Block-level UVM Environment+TB along with Control Plane and Data Plane Agents"""




results_help_text = """Moore.io Results Command
   Parses Simulaton results for a target IP and generates both HTML and Jenkins-compatible XML reports.  These reports
   are output into the simulation directory.
   
Usage:
   mio results IP REPORT_NAME [OPTIONS]
   
Options:
   
Examples:
   mio results my_ip sim_results  # Parse simulation results for 'my_ip' and generate reports under 'sim_results' filenames."""




cov_help_text = """Moore.io Cov(erage) Command
   Merges code and functional coverage data into a single database from which report(s) are generated.  These reports
   are output into the simulation directory.  WARNING: Currenly only supports Vivado.
   
Usage:
   mio cov IP [OPTIONS]
   
Options:
   
Examples:
   mio cov my_ip  # Merge coverage data for 'my_ip' and generate a report."""




doctor_help_text = """Moore.io Doctor Command
   Runs a set of checks to ensure mio installation has what it needs to operate properly.
   
Usage:
   mio doctor [OPTIONS]
   
Options:
   
Examples:
   mio doctor"""




# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import cache
from mio import common
from mio import cfg
from mio import cov
from mio import dox
from mio import results
from mio import sim

import os


def main():
    common.info("Checking EDA tool installations ...")
    check_simulator_executables(cfg.default_simulator)
    common.banner("The Moore.io CLI Client is properly installed and ready to go")


def check_simulator_executables(simulator):
    if simulator == common.simulators_enum.VIVADO:
        if not os.path.exists(cfg.vivado_home):
            common.fatal("Path for vivado executables could not be found " + cfg.vivado_home)
    elif simulator == common.simulators_enum.METRICS:
        if not os.path.exists(cfg.metrics_home):
            common.fatal("Path for metrics executables could not be found " + cfg.metrics_home)
    elif simulator == common.simulators_enum.VCS:
        if not os.path.exists(cfg.vcs_home):
            common.fatal("Path for vcs executables could not be found " + cfg.vcs_home)
    elif simulator == common.simulators_enum.XCELIUM:
        if not os.path.exists(cfg.xcelium_home):
            common.fatal("Path for xcelium executables could not be found " + cfg.xcelium_home)
    elif simulator == common.simulators_enum.QUESTA:
        if not os.path.exists(cfg.questa_home):
            common.fatal("Path for questa executables could not be found " + cfg.questa_home)
    elif simulator == common.simulators_enum.RIVIERA:
        if not os.path.exists(cfg.riviera_home):
            common.fatal("Path for riviera executables could not be found " + cfg.riviera_home)
    return True




#!/usr/bin/python3
########################################################################################################################
# Copyright 2021-2023 Datum Technology Corporation
# All rights reserved.
########################################################################################################################


########################################################################################################################
# IMPORTS
########################################################################################################################
#import gen.ral_typedefs
#import gen.ral_templates
import re
import csv
import jinja2
import tarfile
from jinja2 import Template
from enum import Enum
import os
import sys
from datetime import date
from mio import common
from mio import cfg


########################################################################################################################
# GLOBALS
########################################################################################################################
dbg = False
out_path = ""
default_license = "All rights reserved"
ip_name = ""
name_of_copyright_owner = ""
license = ""
csv_filepath = ""
gen_block_offset = ""
default_gen_block_offset = "0000_0000"
parameters = {}
num_files = 0


########################################################################################################################
# FUNCTIONS
########################################################################################################################
def gen_block():
    common.fatal("Not yet implemented")


def prompt_user_values():
    global out_path
    global csv_filepath
    global ip_name
    global license
    global default_license
    global name_of_copyright_owner
    global parameters
    global gen_block_offset
    global default_gen_block_offset
    
    default_copyright_owner = cfg.org_full_name
    default_vendor          = cfg.org_name
    
    common.info("Moore.io Template Generator: UVM Register Model (v1p0)")
    common.info("******************************************************")
    common.info("The answers to the following questionnaire will be used to generate the code for your new UVM Register Model")
    common.info("")
    
    if out_path == "":
        out_path = common.prompt("Please enter the destination path for this new UVM Register Model (default: '.'):\n").strip()
        if out_path == "":
            out_path = "."
    
    
    name_of_copyright_owner = common.prompt("Please enter the name of the copyright holder or hit RETURN for the default (default is '" + default_copyright_owner + "'):\n").strip()
    if name_of_copyright_owner == "":
        name_of_copyright_owner = default_copyright_owner
    
    name_of_vendor = common.prompt("Please enter a vendor name or hit RETURN for the default (default is '" + default_vendor + "'):\n").strip()
    if name_of_vendor == "":
        name_of_vendor = default_vendor
    
    license = common.prompt("Please enter a usage license or hit RETURN for the default (default is '" + default_license + "'):\n").strip()
    if license == "":
        license = default_license
    
    ip_name = common.prompt("Please enter the name of the UVM Environment to which this Model belongs (ex: 'ctrl'); this name will be used for all types (ex: 'uvme_ctrl_reg_block_c'):\n").lower().strip()
    if ip_name == "":
        common.fatal("ERROR: UVM Environment name cannot be empty.  Exiting.")
    
    csv_filepath = common.prompt("Please enter the path of the input .csv file:\n").lower().strip()
    if csv_filepath == "":
        common.fatal("ERROR: Input .csv file path cannot be empty.  Exiting.")
    
    gen_block_offset = common.prompt("Please enter the base address (default: 0000_0000):\n").lower().strip()
    if gen_block_offset == "":
        gen_block_offset = default_gen_block_offset
    
    parameters = {
        "ip_name"                 : ip_name,
        "year"                    : date.today().year,
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license,
        "vendor"                  : name_of_vendor
    }


def print_end_message():
    global out_path
    global num_files
    common.info(str(num_files) + " files successfully generated into " + out_path)


########################################################################################################################
# ENTRY POINT
########################################################################################################################
def main():
    prompt_user_values()
    gen_block()
    print_end_message()



# Copyright 2021-2023 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import cache
from mio import common
from mio import cfg
from mio import cov
from mio import dox
from mio import results
from mio import sim

import os


def main(ip_str, deep_clean):
    vendor, name = common.parse_dep(ip_str)
    if vendor == "":
        ip = cache.get_anon_ip(name, True)
    else:
        ip = cache.get_ip(vendor, name, True)
    if deep_clean:
        common.banner(f"Deep cleaning IP '{ip.vendor}/{ip.name}'")
        for dep in ip.dependencies:
            dep_ip = cache.get_ip(dep.vendor, dep.target_ip)
            if dep_ip == None:
                common.dbg(f"Cannot find Dependency IP '{dep.vendor}/{dep.target_ip}'")
            else:
                clean_ip(dep_ip)
    else:
        common.banner(f"Cleaning IP '{ip.vendor}/{ip.name}'")
    clean_ip(ip)
    
    if ip.has_dut:
        if ip.dut_ip_type == "fsoc":
            core = cache.get_core(ip.dut_fsoc_full_name)
            if core == None:
                common.fatal(f"Cannot find DUT FuseSoC Core '{ip.dut_fsoc_full_name}'")
            clean_core(core)
        else:
            dut_ip = cache.get_ip(ip.dut.vendor, ip.dut.target_ip)
            if dut_ip == None:
                common.fatal(f"Cannot find DUT IP '{ip.dut.vendor}/{ip.dut.target_ip}'")
            clean_ip(dut_ip)


def clean_core(core):
    common.info(f"Deleting code and compiled code objects for FuseSoC core '{core.sname}'")
    common.remove_dir(cfg.sim_output_dir + '/viv/cmp_wd/@fsoc__' + core.sname)
    common.remove_dir(cfg.sim_output_dir + '/vcs/cmp_wd/@fsoc__' + core.sname)
    common.remove_dir(cfg.sim_output_dir + '/xcl/cmp_wd/@fsoc__' + core.sname)
    common.remove_dir(cfg.sim_output_dir + '/qst/cmp_wd/@fsoc__' + core.sname)
    common.remove_dir(cfg.sim_output_dir + '/riv/cmp_wd/@fsoc__' + core.sname)
    common.remove_dir(cfg.fsoc_dir + "/" + core.sname)
    core.is_installed = False


def clean_ip(ip, no_infos=False):
    if ip.name == "uvm":
        return
    
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir_name = f"{ip.vendor}__{ip.name}"
    
    common.remove_dir(cfg.sim_output_dir + '/viv/cmp_wd/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/mdc/cmp_wd/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/vcs/cmp_wd/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/xcl/cmp_wd/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/qst/cmp_wd/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/riv/cmp_wd/' + ip_dir_name)
    
    common.remove_dir(cfg.sim_output_dir + '/viv/elab_out/single_sim/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/mdc/elab_out/single_sim/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/vcs/elab_out/single_sim/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/xcl/elab_out/single_sim/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/qst/elab_out/single_sim/' + ip_dir_name)
    
    common.remove_dir(cfg.sim_output_dir + '/viv/elab_out/regressions/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/mdc/elab_out/regressions/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/vcs/elab_out/regressions/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/xcl/elab_out/regressions/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/qst/elab_out/regressions/' + ip_dir_name)
    
    if ip.is_local:
        if no_infos:
            common.dbg(f"Deleting compiled code objects for local IP '{ip_str}'")
        else:
            common.info(f"Deleting compiled code objects for local IP '{ip_str}'")
    else:
        if no_infos:
            common.dbg(f"Deleting compiled code objects for external IP '{ip_str}'")
        else:
            common.info(f"Deleting compiled code objects for external IP '{ip_str}'")
    common.remove_dir(cfg.sim_output_dir + '/viv/cmp_out/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/mdc/cmp_out/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/vcs/cmp_out/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/xcl/cmp_out/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/qst/cmp_out/' + ip_dir_name)
    common.remove_dir(cfg.sim_output_dir + '/riv/cmp_out/' + ip_dir_name)
    ip.is_compiled  [common.simulators_enum.VIVADO ] = False
    ip.is_compiled  [common.simulators_enum.METRICS] = False
    ip.is_compiled  [common.simulators_enum.VCS    ] = False
    ip.is_compiled  [common.simulators_enum.XCELIUM] = False
    ip.is_compiled  [common.simulators_enum.QUESTA ] = False
    ip.is_compiled  [common.simulators_enum.RIVIERA] = False
    ip.is_elaborated[common.simulators_enum.VIVADO ] = False
    ip.is_elaborated[common.simulators_enum.METRICS] = False
    ip.is_elaborated[common.simulators_enum.VCS    ] = False
    ip.is_elaborated[common.simulators_enum.XCELIUM] = False
    ip.is_elaborated[common.simulators_enum.QUESTA ] = False
    ip.is_elaborated[common.simulators_enum.RIVIERA] = False
    
